package PrologDB;

import PrologParser.Parser;
import MDELite6.Error;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * DB objects are database instances of DBSchema objects
 * @author don
 */
public class DB extends Parser {

    /**
     * the name of the database; a map of (tableNames,Tables) and the DBSchema
     * that this database instantiates
     */
    private String name;
    private final Map<String, Table> tables;
    private final DBSchema schema;

    // ACCESSORS and INFORMATION
    /**
     * returns name of DB schema that this database instantiates.
     * This method delegates 
     * to its schema object to get its name.
     *
     * @return name of DB schema that this database instantiates 
     */
    public String getSchemaName() {
        return schema.getName();
    }

    /**
     * returns DBSchema that this database instantiates
     * @return DBSchema that this database instantiates
     */
    public DBSchema getSchema() {
        return schema;
    }

    /**
     * returns a list of subtable declarations for this database's schema.
     * this method delegates to the schema object to get its subtables.
     * 
     * @return list of subtable declarations for this database's schema a
     * delegate method
     */
    public List<SubTableSchema> getSubTableSchemas() {
        return schema.getSubTableSchemas();
    }

    /**
     * returns subtable declaration for a given table schema ts;
     * if no such declaration exists, null is returned; this declaration
     * lists the subtables of the given table.
     * 
     * @param ts -- table schema for which subtable declaration is to be returned
     * @return subtable declaration for a given table schema ts
     */
    public SubTableSchema getSubTableSchema(TableSchema ts) {
        for (SubTableSchema s : getSubTableSchemas()) {
            if (s.getSuper().getName().equals(ts.getName())) {
                return s;
            }
        }
        return null;
    }

    /**
     * returns a collection of tables that define this database
     * @return a collection of tables that define this database
     */
    public Collection<Table> getTables() {
        return tables.values();
    }

    /**
     * returns the TableSchema of the table with given name; null if table is
     * not found
     *
     * @param name name of desired table
     * @return either TableSchema of this table or null
     */
    public TableSchema getTableSchema(String name) {
        return schema.getTableSchema(name);
    }

    /**
     * returns the TableSchema of the table with given name; Error
     * is thrown otherwise
     *
     * @param name name of desired table
     * @return either TableSchema of this table or null
     */
    public TableSchema getTableSchemaEH(String name) {
        return schema.getTableSchemaEH(name);
    }
    
    /**
     * returns the name of this database
     * 
     * @return name of this database
     */
    public String getName() {
        return name;
    }

    /**
     * the full name of a database is "databaseName"."schemaName"."pl"; this is
     * the assumed or standard name for a prolog database file.
     *
     * @return returns full name of schema
     */
    public String getFullName() {
        return String.format("%s.%s.pl", name, getSchema().getName());
    }

    /**
     * Returns true if there is a table with tableName in the database
     *
     * @param tableName of table
     * @return true if table is present, false otherwise
     */
    public boolean containsTable(String tableName) {
        return tables.containsKey(tableName);
    }

    // CONSTRUCTORS
    /**
     * Create a prolog database with a given name that instantiates db schema
     *
     * @param name of the database
     * @param schema of the database
     */
    public DB(String name, DBSchema schema) {
        this.name = name;
        this.schema = schema;
        tables = new HashMap<>();
        for (TableSchema t : schema.getTableSchemas()) {
            Table tab = new Table(t);
            tables.put(t.getName(), tab);
            tab.parent = this;
        }
    }

    /**
     * reads and parses a prolog database from File infile. may throw
     * parseExceptions or Errors, which terminates processing. Errors are
     * reported to System.err.
     *
     * @param infile the File of the Prolog database
     * @return Prolog database object of type DB
     */
    public static DB readDataBase(File infile) {
        return DB.readDataBase(infile, System.err);
    }

    /**
     * reads and parses a prolog database from given string filename. may throw
     * parseExceptions or Errors, which terminates processing. Errors are
     * reported to System.err.
     *
     * @param localFileName is the name of the file of the Prolog database
     * @return Prolog database object of type DB
     */
    public static DB readDataBase(String localFileName) {
        DB.dbCheck(localFileName);

            File in = new File(localFileName);
            if (!in.exists())
                throw new Error(localFileName + " does not exist");
            return readDataBase(in);
    }

    /**
     * reads and parses a prolog database from given string filename. may throw
     * parseExceptions or Errors, which terminates processing. Errors are
     * reported to Printstream out.
     *
     * @param infile is the File of the Prolog database
     * @param out is the PrintStream (e.g., System.out) to report errors.
     * @return Prolog database object of type DB
     */
    public static DB readDataBase(File infile, PrintStream out) {
        DB db;
        LineNumberReader br;
        String filename = infile.getName();
        String[] parts = filename.split("/");
        String[] subparts = parts[parts.length - 1].split("\\.");
        if (!subparts[2].equals("pl")) {
            throw new Error("file name " + filename + " not in <dbname>.<schemaname>.pl");
        }
        String dbname = subparts[0];
        // read the file in the first pass: get its schema
        DBSchema dbs = DBSchema.readSchema(infile, out);

        db = new DB(dbname, dbs);
        String line;
        int lineno = 1;
        try {
            // Step 1: Read infile by line
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(infile)));

            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("") || line.startsWith("tuple(") || line.startsWith("dbase(") || line.contains(":-") || line.startsWith("table(")
                        || line.startsWith("subtable(")
                        || line.startsWith("/*") || line.startsWith("%")) {
                    continue;
                }

                parseTupleDecl(line, br.getLineNumber(), db); // if not tuple line, ignore
            }
            br.close();
        } catch (Exception e) {
            String err = "in file " + infile.getName() + " : " + e.getMessage();
            throw new Error(err);
        }
        return db;
    }


    // UTILITIES
    /**
     * assigns a new name to this database
     *
     * @param name new name
     */
    public void rename(String name) {
        this.name = name;
    }

    /**
     * asserts that a file pathName ends in ".pl";
     * Error is thrown otherwise
     * 
     * @param pathName -- filename to verify
     */
    public static void dbCheck(String pathName) {
        if (!pathName.endsWith(".pl")) {
            throw new Error("prolog database file does not end in '.pl': "+ pathName);
        }
    }

    // SEARCH UTILITIES
//    /**
//     * return the list of ALL tuples from the table with name tableName; there
//     * are no qualifications. Tuples from all subtables of this table will be
//     * returned AND trimmed of unnecessary columns.
//     *
//     * @param tableName name of table
//     * @return list of all column-projected tuples in table and its subtables
//     */
//    public TupleList tuples(String tableName) {
//        Table tbl = getTableEH(tableName);
//        return tbl.tuples();
//    }

    /**
     * returns the Table object for the table with name tableName;
     * null is returned if table does not exist
     *
     * @param tableName -- name of table to return
     * @return Table object or null
     */
    public Table getTable(String tableName) {
        return tables.get(tableName);
    }

    /**
     * error handler of getTable(String); return the Table object for the table
 with name tableName; Error is thrown if no such table exists.
     *
     * @param tableName -- name of table to return
     * @return Table object or throws Error
     */
    public Table getTableEH(String tableName) {
        Table t = tables.get(tableName);
        if (t == null) {
            throw new Error("cannot find table " + tableName + " in database " + schema.getName());
        }
        return t;
    }

    // PRINT
    /**
     * print in standard way the contents of a Prolog database to PrintStream out
     *
     * @param out destination of printing
     */
    public void print(PrintStream out) {
        out.format("%s\n", schema.printDBaseDecl());

        for (TableSchema s : schema.getTableSchemas()) {
            Table t = tables.get(s.getName());
            t.print(out);
        }

        for (SubTableSchema sts : schema.getSubTableSchemas()) {
            sts.print(out);
        }
    }
}
