package PrologDB;

import PrologParser.Parser;
import MDELite6.Error;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

/**
 * A DBSchema object defines the schema of a database
 *
 */
public class DBSchema extends Parser {

    /**
     * name of database schema
     */
    private String name;

    /**
     * list of table tableSchemas, one per table in the database schema
     *
     */
    private final LinkedList<TableSchema> tableSchemas;

    /**
     * list of subtable declarations, which encodes a forest of inheritance
     * hierarchies
     */
    private LinkedList<SubTableSchema> subtables;

    /**
     * Has the schema been flattened? a schema is flattened when associated with
     * a DB (prolog database object); see flatten() operation below.
     */
    private boolean flattened;

    // ACCESSORS AND INFOMATION
    /**
     * return the name of the schema
     *
     * @return the name of the schema
     */
    public String getName() {
        return name;
    }

    /**
     * full (file) name of a schema is (name).schema.pl
     *
     * @return manufactured file name of schema
     */
    public String getFullName() {
        return name + ".schema.pl";
    }

    /**
     * return list of TableSchemas of this database schema
     *
     * @return list of TableSchemas of this database schema
     */
    public LinkedList<TableSchema> getTableSchemas() {
        return tableSchemas;
    }

    /**
     * return list of SubTableSchemas of this database schema
     *
     * @return list of SubTableSchemas of this database schema
     */
    public LinkedList<SubTableSchema> getSubTableSchemas() {
        return subtables;
    }

    /**
     * returns subtable object for given tableName, null if none exists
     *
     * @param tableName -- name of table whose subtables are to be returned
     * @return subtable object for given tableName
     */
    public SubTableSchema getSubTableSchema(String tableName) {
        for (SubTableSchema s : subtables) {
            if (s.getName().equals(tableName)) {
                return s;
            }
        }
        return null;
    }

    /**
     * returns subtable object for given tableName; throws Error if no
     * such tableName exists
     * 
     * @param tableName -- name of table whose subtables are to be returned
     * @return subtable object for given tableName
     */
    public SubTableSchema getSubTableSchemaEH(String tableName) {
        SubTableSchema result = getSubTableSchema(tableName);
        if (result == null) {
            throw new Error("schema " + tableName + " missing from database");
        }
        return result;
    }

    /**
     * returns subtable object for given tableSchema ts, null if none exists
     *
     * @param ts -- TableSchema object for which subtable declaration is to be returned
     * @return subtable object for given tableSchema ts
     */
    public SubTableSchema getSubTableSchema(TableSchema ts) {
        return getSubTableSchema(ts.getName());
    }

    /**
     * returns subtable object for given tableSchema ts; throws Error
     * if no such ts exists
     *
     * @param ts -- name of tableSchema whose subtables are to be returned
     * @return subtable object for given tableSchema ts
     */
    public SubTableSchema getSubTableSchemaEH(TableSchema ts) {
        SubTableSchema result = getSubTableSchema(ts);
        if (result == null) {
            throw new Error("schema " + ts.getName() + " missing from database");
        }
        return result;
    }

    /**
     * return number of TableSchemas (equivalently the number of Tables in a
     * schema instance) in this dbschema
     *
     * @return number of TableSchemas in this db schema
     */
    public int size() {
        return tableSchemas.size();
    }

    /**
     *
     * @param list
     * @return
     */
    private SubTableSchema dominant(LinkedList<SubTableSchema> list) {
        boolean foundit = true;
        SubTableSchema last = null;

        for (SubTableSchema outer : list) {
            last = outer;
            for (SubTableSchema inner : list) {
                if (inner.contains(outer.getSuper())) {
                    foundit = false;
                    break;
                }
            }
            if (foundit) {
                return outer;
            }
            foundit = true;
        }
        throw new Error("table " + last.getName() + " participates in inheritance cycle");
    }

    // CONSTRUCTORS
    /**
     * create DBSchema that has name and a list of TableSchemas can
     * incrementally add subtable tableSchemas subsequently -- before schema is
     * instantiated as a database
     *
     * @param name of DBSchema
     * @param schemas comma-separated list of table tableSchemas
     */
    public DBSchema(String name, TableSchema... schemas) {
        this.name = name;
        this.tableSchemas = new LinkedList<>();
        this.tableSchemas.addAll(Arrays.asList(schemas));
        this.subtables = new LinkedList<>();
        this.flattened = false;
    }

    /**
     * This constructor is used for the incremental construction of database
     * tableSchemas. First provide a name of the schema, then incrementally add
     * tableSchemas per table and then subtables. By default, schema is not
     * flattened.
     *
     * @param name -- name given to the DBSchema
     */
    public DBSchema(String name) {
        this.name = name;
        this.tableSchemas = new LinkedList<>();
        subtables = new LinkedList<>();
        flattened = false;
    }

    /**
     * copy a database schema -- please do not alter existing table definitions
     * and subtable definitions. can add more tables, subtable definitions.
     * Copied schema is flattened.
     *
     * @return copied schema
     */
    public DBSchema copy() {
        DBSchema s = new DBSchema(name);
        s.tableSchemas.addAll(tableSchemas);
        s.subtables.addAll(subtables);
        s.flattened = flattened;
        return s;
    }

    // can read .ooschema and .schema files
    /**
     * reads .ooschema and .schema files; throws Error when an error is
     * encountered. Returns DBSchema declaration in file; may flatten it if
     * .ooschema is read.
     *
     * @param schemafile -- name of schema file
     * @return DBSchema declaration of file
     */
    public static DBSchema readSchema(String schemafile) {
        File in = new File(schemafile);
        if (!in.exists()) {
            throw new Error(schemafile + " does not exist");
        }
        return readSchema(in, System.err);
    }

    /**
     * reads prolog schema from File schemafile; errors are reported to out
     *
     * can read .ooschema and .schema files
     *
     * @param schemafile Java File of schema file to read (can also be a
     * database file, as it too embeds schema information)
     * @param out PrintStream to report errors
     * @return DBSchema object
     */
    public static DBSchema readSchema(File schemafile, PrintStream out) {
        DBSchema dbs = null;
        List<String> tableNames = new ArrayList<>();

        File f;

        if (!schemafile.exists()) {
            throw new Error(schemafile.getName() + " does not exist");
        }

        try {
            LineNumberReader br = new LineNumberReader(new InputStreamReader(new FileInputStream(schemafile)));
            String line;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("") ||  line.startsWith("%")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                try {
                    if (line.startsWith("dbase(")) {
                        if (dbs != null) {
                            throw new Error("more than one dbase() declaration in file " + schemafile.getName());
                        }
                        dbs = parseDBaseDecl(line, lineno, tableNames);
                    } else if (line.startsWith("table")) {
                        parseTableDecl(line, lineno, dbs);
                    } else if (line.startsWith("subtable")) {
                        parseSubTableDecl(line, lineno, dbs);
                    }
                    // ignore everything else
                } catch (Exception e) {
                    String err = String.format("unable to parse table on line %d in %s\n%s\n",
                            br.getLineNumber(), schemafile.getName(), e.getMessage());
                    throw new Error(err);
                }
            }
            br.close();
            // we have parsed a database schema file
            // Let's make sure that the tables declared match what is in the dbase declaration

            if (dbs == null) {
                throw new Error(schemafile.getName() + " is not a valid prolog database schema!");
            }

            dbs.flattened = schemafile.getName().endsWith(".schema.pl");

            for (String s : tableNames) {
                dbs.getTableSchemaEH(s); // throw error if not found
            }
            for (TableSchema ts : dbs.getTableSchemas()) {
                if (!tableNames.contains(ts.getName())) {
                    throw new Error("table " + ts.getName() + " not defined in dbase decl");
                }
            }
        } catch (IOException e) {
            throw new Error(e.getMessage());
        }
        return dbs;

        // finally, depending on who is calling this, the flatten attribute is to be set.
    }

    // CONSTRUCTOR UTILITIES
    /**
     * add table schema tableschema to 'this' database schema
     *
     * @param tableschema to add to 'this' database schema
     */
    public void addTableSchema(TableSchema tableschema) {
        String tableSchemaName = tableschema.getName();
        if (getTableSchema(tableSchemaName) != null) {
            throw new Error("multiple table declarations for " + tableSchemaName);  // throw error if not present
        }
        tableSchemas.add(tableschema);
        tableschema.parent = this;
    }

    /**
     * add subtable schema sts to 'this' database schema
     *
     * @param sts subtableschema to add to 'this' database
     */
    public void addSubTableSchema(SubTableSchema sts) {
        String tableName = sts.getName();
        if (getSubTableSchema(tableName) != null) {
            throw new Error("multiple subtable declarations for " + tableName);
        }
        subtables.add(sts);
        sts.parent = this;
    }

    // SEARCH UTILITIES
    /**
     * return the TableSchema with name tableName
     *
     * @param tableName of schema to return
     * @return the TableSchema with name tableName, null if not found
     */
    public TableSchema getTableSchema(String tableName) {
        for (TableSchema ts : tableSchemas) {
            if (ts.getName().equals(tableName)) {
                return ts;
            }
        }
        return null;
    }

    /**
     * return the TableSchema with name tableName
     *
     * @param tableName of schema to return
     * @return the TableSchema with name tableName, throw Error if not foundd
     */
    public TableSchema getTableSchemaEH(String tableName) { // built-in error handling
        TableSchema ts = getTableSchema(tableName);
        if (ts == null) {
            throw new Error("no declaration for table " + tableName);
        }
        return ts;
    }

    // UTILITIES
    /**
     * asserts that schemaFilename must end with ".schema.pl"; throws
     * Error otherwise
     *
     * @param schemaFileName -- file name to qualify
     */
    public static void schemaCheck(String schemaFileName) {
        if (!schemaFileName.endsWith(".schema.pl")) {
            throw new Error(String.format("prolog schema file does not end in '.schema.pl': %s\n", schemaFileName));
        }
    }

    /**
     * asserts that schemaFileName, which could be an .ooschema or .schema, ends
     * with the appropriate ".schema.pl" or ".ooschema.pl" extensions, otherwise
     * a Error is thrown
     *
     * @param schemaFileName -- name of schema file to check
     */
    public static void anyschemaCheck(String schemaFileName) {
        if (schemaFileName.endsWith(".schema.pl") || schemaFileName.endsWith(".ooschema.pl")) {
            return;
        }
        throw new Error(String.format("prolog schema file does not end in '.schema.pl' or '.ooschema.pl': %s\n",
                schemaFileName));
    }

    /**
     * two DBSchemas are equal iff (a) they have the same name, (b) they have
     * the same set of table schemas, and (c) they have the same set of
     * subtableschemas; if so true is returned.
     *
     * @param dbschema -- the other dbschema 
     * @return true if both DBSchemas are equal.
     */
    public boolean equals(DBSchema dbschema) {
        if (!dbschema.getName().equals(getName())) {
            return false;   // schema names are not the same
        }
        if (dbschema.size() != size()) {
            return false;  // # of table definitions are not the same
        }
        for (TableSchema s : dbschema.getTableSchemas()) {
            TableSchema r = getTableSchema(s.getName());
            if (r == null || !r.equals(s)) {
                return false; // tables don't match
            }
        }

        if (subtables.size() != dbschema.subtables.size()) {
            return false;
        }

        for (SubTableSchema sts : dbschema.subtables) {
            SubTableSchema s2 = getSubTableSchema(sts.getName());
            if (!sts.equals(s2)) {
                return false; // have different subtable entries
            }
        }
        return true;  // they're the same
    }

    /**
     * two DBSchemas are equal iff (a) they have the same name, (b) they have
     * the same set of table schemas, and (c) they have the same set of sub
     * table schemas; if so this method returns. Otherwise a Error is
     * thrown with an explanation
     *
     * @param dbschema.
     */
    public void equalsEH(DBSchema dbschema) {
        if (!dbschema.getName().equals(getName())) {
            throw new Error("schema names are not the same");
        }
        if (dbschema.size() != size()) {
            throw new Error("# of table definitions are not the same");
        }
        for (TableSchema s : dbschema.getTableSchemas()) {
            TableSchema r = getTableSchema(s.getName());
            if (r == null) {
                throw new Error("table schema " + s.getName() + " is missing");
            }
            r.equalsEH(s);
        }

        if (subtables.size() != dbschema.subtables.size()) {
            throw new Error("# of subtable definitions are not the same");
        }

        for (SubTableSchema sts : dbschema.subtables) {
            SubTableSchema s2 = getSubTableSchema(sts.getName());
            sts.equalsEH(s2);
        }
    }

    /**
     * rename DBschema
     *
     * @param name new DBSchema name
     */
    public void rename(String name) {
        this.name = name;
    }

    /**
     * propagate attributes of superTables to subTables (operation is performed
     * on schema defs)
     */
    public void flatten() {
        // Step 1: must order subtable list by dominance (parent dominates a child)
        LinkedList<SubTableSchema> list = subtables;
        subtables = new LinkedList<>();
        while (!list.isEmpty()) {
            SubTableSchema next = dominant(list);
            list.remove(next);
            subtables.add(next);
        }

        // Step 2: propagate parent attributes to subattributes
        for (SubTableSchema st : subtables) {
            st.flatten();
        }

        flattened = true;
    }

    // PRINT
    String printDBaseDecl() {
        String db = "dbase(" + name + ",[";
        String comma = "";
        for (TableSchema ts : tableSchemas) {
            db = db + comma + ts.getName();
            comma = ",";
        }
        return db + "]).\n";
    }

    /**
     * converts DBSchema to a string. Works for ooschema and schema files
     *
     * @return print string of DBSchema
     */
    public String toString() {
        String result = String.format("%s\n", printDBaseDecl());
        for (TableSchema ts : tableSchemas) {
            result = result + ts.toString();
        }
        if (!subtables.isEmpty()) {
            result = result + "\n";
            for (SubTableSchema st : subtables) {
                result = result + st.toString();
            }
        }
        return result;
    }

    /**
     * print the database schema definition to PrintStream out; works for
     * ooschemas and schema files
     *
     * @param out PrintStream destination
     */
    public void print(PrintStream out) {
        out.print(this.toString());
    }

}
