package PrologDB;

import MDELite6.Error;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

/**
 * a Table is an instance of a TableSchema. Both share the same names.
 */
public class Table {

    /**
     * each table points to its tablesSchema "schema"
     */
    protected final TableSchema schema;

    /**
     * each table has a TupleList of "tuples"
     */
    private TupleList tuples;

    /**
     * each table has a list of its subtables; this list is initialized by
     * Table.getSubTables upon first reference
     */
    private LinkedList<Table> subtables;  // initialized by Table.getSubTables()

    /**
     * each table points to its DBSchema "parent". Normally, the value of this
     * field is set by a call to a DB.constructor. However, tables that are
     * created that do not belong to a DBSchema have this field null.
     */
    DB parent = null;

    // CONSTRUCTORS
    /**
     * create an empty table for TableSchema schema
     *
     * @param schema -- the schema to which the table should conform
     */
    public Table(TableSchema schema) {
        this.schema = schema;
        tuples = new TupleList();
        subtables = null;
    }

    /**
     * this private constructor is used for self joins, where tuples are copied
     * from one schema to another.
     *
     * Precondition: schemas of TupleList tab and schema have identically
     * defined columns. Names of schemas could be different.
     *
     * @param schema -- schema of table to create
     * @param tab -- table tuples to copy into the new table
     */
    private Table(TableSchema schema, Table tab) {
        if (schema.isDetached()) {
            this.schema = schema;
        } else {
            this.schema = schema.CopyForSelfJoins(schema.getName());
        }
        tuples = tab.tuples.copy(schema);
        subtables = null;
    }

    /**
     * this method copies a table t for use in self-joins; the original table t
     * is copied and renamed to copyTableName
     *
     * @param copyTableName -- postfix name distinguisher
     * @return copied table
     */
    public Table copyForSelfJoins(String copyTableName) {
        TableSchema renamedSchema = this.schema.CopyForSelfJoins(copyTableName);
        Table tnew = new Table(renamedSchema, this);
        return tnew;
    }

    /**
     * reads and parses a prolog table from given string filename. may throw
     * parseExceptions or Errors, which terminates processing. Errors are
     * reported to PrintStream out.
     *
     * @param tableFilePath is the filename of the Prolog database
     * @return Table of tableFilePath
     */
    public static Table readTable(String tableFilePath) {
        if (tableFilePath.endsWith(".csv")) {
            return readCSVTable(tableFilePath);
        }

        LineNumberReader br;
        File infile;
        if (!tableFilePath.endsWith("pl")) {
            throw new Error("file name " + tableFilePath + " does not end in '.pl'");
        }

        // read the file in the first pass: get its schema
        try {
            infile = new File(tableFilePath);
        } catch (Exception e) {
            throw new Error(e.getMessage());
        }
        TableSchema ts = TableSchema.readSchema(infile);

        // second pass - read the table itself
        Table t = new Table(ts);
        String line;
        try {
            // Step 1: Read infile by line
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(infile)));

            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("") || line.startsWith("table(") || line.startsWith("%")) {
                    continue;
                }

                Tuple tuple = Tuple.parseTupleDecl(line, br.getLineNumber(), ts); // if not tuple line, ignore
                t.add(tuple);
            }
            br.close();
        } catch (Exception e) {
            String err = "in file " + infile.getName() + " : " + e.getMessage();
            throw new Error(err);
        }
        return t;
    }

    /**
     * read a csv file whose path-name is tablePathName 1st line of this file
     * defines the table schema attributes (all are assumed "quoted") lines 2-$
     * of this fine each define a tuple of this table returns Table of csv file
     *
     * @param tablePathName -- pathname to csv file
     * @return Table of the read csv file
     */
    private static Table readCSVTable(String tablePathName) {
        TableSchema ts;
        Table table = null;
        LineNumberReader br = null;
        try {
            // Step 1: read the first line of the csv file and interpret it as a
            //         schema declaration
            String name = getCSVName(tablePathName);
            File schemafile = new File(tablePathName);
            br = new LineNumberReader(new InputStreamReader(new FileInputStream(schemafile)));
            String line = br.readLine();
            String[] splt = line.split(",");
            ts = new TableSchema(name);
            for (String i : splt) {
                ts.addColumn(new Column(i.trim(), true));
            }

            // Step 2: subsequent lines define a single tuple, which is
            //         added to the newly created table "table"
            table = new Table(ts);
            while ((line = br.readLine()) != null) {
                line = line + " ,";
                String[] spl = line.split(",");
                String tmp = spl[spl.length - 1].trim();
                spl[spl.length - 1] = tmp;
                Tuple tuple = new Tuple(ts);
                tuple.addColumnValuesEH(spl);
                table.add(tuple);
            }
            br.close();
        } catch (Exception e) {
            String lnum = "";
            if (br != null) 
                lnum = "" + br.getLineNumber();
            throw new Error(e.getMessage() + " on line " + lnum);
        }
        return table;
    }

    // RELATIONAL OPERATIONS
    /**
     * transform a table to an object of type S
     *
     * @param <S> -- return type, should not be Table or subclass of Table
     * @param f -- a function that maps a Table to S
     * @return -- object of type S
     */
    public <S> S map(Function<Table, S> f) {
        return f.apply(this);
    }

    /**
     * return a Table from a given Table that satisfies predicate p
     *
     * @param p -- predicate to satisfy
     * @return predicate reduced TupleList
     */
    public Table filter(Predicate<Tuple> p) {
        // Step 1: create empty table with same schema as t
        Table shorterTable = new Table(schema);

        // Step 2: dup over every tuple in t that satisfies p
        this.stream().filter(p).forEach(tt -> shorterTable.add(tt));

        // Step 3: output row-reduced table
        return shorterTable;
    }

    /**
     * projects a Table to the set of columns of a super table schema if given
     * superTableSchema does not match any tuple in the list an Error is thrown
     *
     * @param superTableSchema -- schema of supertable to project to
     * @return column-projected Table
     */
    public Table project(TableSchema superTableSchema) {
        Table shorterTable = new Table(superTableSchema);
        for (Tuple t : tuples) {
            shorterTable.add(t.project(superTableSchema));
        }
        return shorterTable;
    }

    /**
     * project Table to a list of cols
     *
     * @param cols - List of columns
     * @return projected Table
     */
    public Table project(List<Column> cols) {
        TableSchema newSchema = new TableSchema(schema.getName(), cols);
        Table shorterTable = new Table(newSchema);
        for (Tuple o : tuples) {
            Tuple tuple = new Tuple(newSchema);

            for (Column c : cols) {
                String name = c.getName();
                String value = o.get(name);
                tuple.addColumnValue(name, value);
            }
            shorterTable.add(tuple);
        }
        return shorterTable;
    }

    /**
     * project Table to an array of column names. precondition: each name must
     * correspond to an existing column in schema
     *
     * @param colNames - List of column names
     * @return projected TupleList
     */
    public Table project(String... colNames) {
        TableSchema newts = new TableSchema(schema.getName());
        for (String c : colNames) {
            Column k = schema.getColumnEH(c);
            newts.addColumn(k);
        }

        Table shorterTable = new Table(newts);
        for (Tuple o : tuples) {
            Tuple tuple = new Tuple(newts);

            for (String name : colNames) {
                String value = o.get(name);
                tuple.addColumnValue(name, value);
            }
            shorterTable.add(tuple);
        }
        return shorterTable;
    }
    
    /**
     * this is a table constructor that takes the equijoin of innerTable on the
     * innerJoinKey (column name) with outerTable on the outerJoinKey (column
     * name). The crossproduct of both schemas is formed, and the joined table
     * is returned.
     *
     * @param innerTable -- inner table of join
     * @param innerJoinKey -- join key of inner table
     * @param outerTable -- outer table of join
     * @param outerJoinKey -- join key of outer table
     * @return equijoin of innerTable and outerTable
     */
    public static Table join(Table innerTable, String innerJoinKey, Table outerTable, String outerJoinKey) {
        TableSchema X = innerTable.getSchema().crossProduct(outerTable.getSchema());
        Table x = new Table(X);
        for (Tuple in : innerTable.tuples) {
            for (Tuple out : outerTable.tuples) {
                if (in.is(innerJoinKey, out.get(outerJoinKey))) {
                    Tuple t = new Tuple(X);
                    t.addColumnValues(in);
                    t.addColumnValues(out);
                    x.add(t);
                }
            }
        }
        return x;
    }

    /**
     * dup Table by copying each tuple and returning the copied Table. Unchecked
     * Precondition: this.tableSchema must have an identical set of columns as
     * tableSchema. Only the names of their schemas may be different
     *
     * @param tableSchema -- table schema of the list
     * @return copied tuple list
     */
    public Table copy(TableSchema tableSchema) {
        Table dup = new Table(tableSchema);
        stream().forEach((t) -> {
            dup.add(t.copy(tableSchema));
        });
        return dup;
    }

    /**
     * sort 'this' Table on column with columnName
     *
     * @param columnName -- name of (string) column to sort
     * @return columnName sorted TupleList
     */
    public Table sort(String columnName) {
        schema.hasColumnEH(columnName);
        Comparator<Tuple> comp = (Tuple a, Tuple b) -> {
            return a.get(columnName).compareTo(b.get(columnName));
        };

        Collections.sort(tuples, comp);
        return this;
    }

    // CONSTRUCTOR UTILITIES
    /**
     * adds tuple t to 'this' table; throws error the schema of the tuple does
     * not match the schema of the table to which it is being added Untested
     * precondition: that the schema of the tuple conforms to the schema of the
     * table
     *
     * @param t is tuple to add
     * @return updated Table ('this');
     */
    public Table add(Tuple t) {
        tuples.add(t);
        return this;
    }
    
    /** finds tuple with (key,value) pair and deletes the first
     * 
     * @param key  -- column
     * @param value -- column value
     */
    public void delete(String key, String value) {
        for (Tuple t :  tuples) {
            if (t.is(key,value)) {
                tuples.remove(t);
                return;
            }
        }
        throw new Error(String.format("attempt to delete tuple with (%s = '%s'), not found in table %s\n",
              key,value, getName()));
    }

    // ACCESSORS and INFORMATION
    /**
     * returns the DBSchema of this table. If there is no such DBSchema null is
     * returned.
     *
     * @return DBSchema of this table
     */
    public DBSchema getDBSchema() {
        return schema.parent;
    }

    /**
     * returns the DBSchema of this table. if there is no DBSchema (i.e., it is
     * null) an Error is thrown
     *
     * @return DBSchema of this table
     *
     */
    public DBSchema getDBSchemaEH() {
        if (schema.parent == null) {
            throw new Error("Table " + getName() + " not a member of any database");
        }
        return schema.parent;
    }

    /**
     *
     * returns name of table, which always equals the name of its table schema
     *
     * @return name of table
     */
    public String getName() {
        return schema.getName();
    }

    /**
     *
     * returns schema of table
     *
     * @return TableSchema of 'this' table
     */
    public TableSchema getSchema() {
        return schema;
    }

    /**
     * returns list of columns of the table; delegates to getColumns of
     * TableSchema
     *
     * @return list of columns of this table
     */
    public List<Column> getColumns() {
        return schema.getColumns();
    }

    /**
     * return list of Tuples of this table and all of its subtables no
     * projection is done to ensure that all tuples have exactly and only the
     * columns of its parents.
     *
     * @return list of tuples of this table and all of its subtables
     */
    public TupleList tuples() {
        TupleList tl = new TupleList();

        if (parent == null) {
            // table that is computed, does not belong to a schema
            return tuplesLocal();
        }

        List<Table> tablist = getSubTables();
        for (Table t : tablist) {
            tl.addAll(t.tuplesLocal());
        }
        return tl;
    }

    /**
     * return only the tuples of this table, and none of its subtables Generally
 not used; tuples() is preferred
     *
     * @return list of tuples of only 'this' table
     */
    public TupleList tuplesLocal() {
        return tuples;
    }
    
    /**
     * return first tuple in table that satisfies predicate p
     * @param p - filter or query, usually over a key field
     * @return null if there is no such tuple, or the first tuple
     */
    public Tuple getFirst(Predicate<Tuple> p) {
        for (Tuple t : tuples) {
            if (p.test(t)) 
                return t;
        }
        return null;
    }
     
    /**
     * return first tuple in table that satisfies predicate p
     * @param p - filter or query, usually over a key field
     * @return tuple ; throws error if not found
     */
    public Tuple getFirstEH(Predicate<Tuple> p) {
       Tuple t = getFirst(p);
       if (t == null) {
           throw new Error("no tuple in table " + getName() + " satisfies given predicate");
       }
       return t;
    }

    /**
     *
     * returns number of tuples in this table
     *
     * @return number of tuples in this table
     */
    public int size() {
        return tuples.size();
    }

    /**
     * returns list of table + its subtables to search note: this method is
     * really executed once. By doing so, the "subtables" member of "this" is
     * set to the list of previously computed subtables.
     *
     * @return list of all tables for the given table, which includes the table
     * itself
     */
    public List<Table> getSubTables() {

        // Step 0: if we have computed this list before, return it
        if (subtables != null) {
            return subtables;
        }

        // Step 1: initialize list with root of inheritance hierarchy
        subtables = new LinkedList<>();
        subtables.add(this);

        // Step 2: find a subtable schema for this table schema
        //         if there is none, we're done
        SubTableSchema sts = parent.getSubTableSchema(schema);
        if (sts == null) {
            return subtables;
        }

        // Step 3: add the list of subtable names, recursively
        for (TableSchema ts : sts.getSubTableSchemas()) {
            Table t = parent.getTableEH(ts.getName());
            subtables.addAll(t.getSubTables());
        }
        return subtables;
    }

    /**
     * returns the name of a csv file, sans ".csv" and any directory information
     * linux example: "a/b/c.csv" returns "c" windoze example: "a\b\c.csv"
     * returns "c"
     *
     * @param pathName is the path-name to a csv file
     * @return cvs file name
     */
    public static String getCSVName(String pathName) {
        if (!pathName.endsWith(".csv")) {
            throw new Error(pathName + " does not end in .csv");
        }
        String name = pathName.replace(".csv", "");
        int slash = name.lastIndexOf('/');
        int bslash = name.lastIndexOf('\\');
        if (slash == -1 && bslash == -1) {
            return name;
        }
        if (slash != -1 && bslash != -1) {
            throw new Error("file name has / and \\ in it!");
        }
        if (slash != -1) {
            return name.substring(slash + 1);
        } else {
            return name.substring(bslash + 1);
        }
    }

    // PUBLIC UTILITIES AND PRINT
    /**
     * print table to PrintStream out
     *
     * @param out -- PrintStream to output print
     */
    public void print(PrintStream out) {
        if (schema == null) {
            throw new Error("schema not set in table " + getName());
        }
        schema.print(out);
        tuples.stream().forEach(t -> t.print(out));
        out.format("\n");
    }

    /**
     * general writeTable file utility -- if filename ends in ".csv"
     * then the table will be written in CSV format.
     * @param tablePathName -- file path name
     */
    public void writeTable(String tablePathName) {
        try {
            PrintStream out = new PrintStream(tablePathName);
            if (tablePathName.endsWith(".csv")) {
                writeCSVTable(out);
            } else {
                print(out);
            }
        } catch (Exception e) {
            System.err.println(e.getMessage());
        }
    }

    /**
     * writeTable table in CSV form to PrintStream out
     *
     * @param out -- PrintStream to writeTable CSV table
     */
    private void writeCSVTable(PrintStream out) {
        TableSchema ts = this.getSchema();
        String comma = "";
        for (Column c : ts.getColumns()) {
            out.format("%s%s", comma, c.getName());
            comma = ",";
        }
        out.format("\n");
        for (Tuple tup : tuples) {
            comma = "";
            for (Column k : tup.getColumns()) {
                String val = tup.get(k.getName()).trim();
                out.format("%s%s", comma, val);
                comma = ",";
            }
            out.format("\n");
        }
    }

    /**
     * convert table into a stream of Tuples
     *
     * @return stream of table tuples
     */
    public Stream<Tuple> stream() {
        return tuples().stream();
    }
}
