package PrologDB;

import MDELite6.Error;
import MDELite6.ParseException;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.LineNumberReader;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;
import static PrologParser.Parser.parseTableDecl;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;

/**
 * a TableSchema defines the name of a table and its columns
 */
public class TableSchema {

    /**
     * schemaName of table schema
     */
    private final String name;

    /**
     * list of its columns
     */
    private final List<Column> columns;

    /**
     * DBSchema to which this table belongs. if null, tableSchema is not part of
     * any DBSChema
     */
    DBSchema parent = null;

    // ACCESSORS AND INFORMATION
    /**
     * returns schemaName of table schema
     *
     * @return name of this tableSchema
     */
    public String getName() {
        return name;
    }

    /**
     * returns the ith Column definition of the table; throws RuntimeError if i
     * is out of bounds. Should not be exported from this package, but necessary
     * for parser (sigh)
     *
     * @param i column index
     *
     * @return returns the ith Column definition of 'this' table
     */
    public Column getIndex(int i) {
        return columns.get(i);
    }

    /**
     * returns list of columns of a TableSchema
     *
     * @return list of columns of 'this' tableSchema
     */
    public List<Column> getColumns() {
        return columns;
    }

    /**
     * quickLookup is a cache for verifying legitimate field names quickly
     */
    private HashSet<String> quickLookup = new HashSet<>();

    /**
     * does schema have a column with columnName?
     *
     * @param columnName -- column name to verify existence
     */
    void hasColumnEH(String columnName) {
        if (quickLookup.contains(columnName)) {
            return;
        }
        for (Column c : columns) {
            if (c.getName().equals(columnName)) {
                quickLookup.add(columnName);
                return;
            }
        }
        throw new Error("unrecognized column " + columnName);
    }
    
    /**
     * does schema NOT have a column with columnName?
     *
     * @param columnName -- column name to verify nonexistence
     */
    void doesNotHaveColumnEH(String columnName) {
        if (!quickLookup.contains(columnName)) {
            return;
        }
        for (Column c : columns) {
            if (c.getName().equals(columnName)) {
                throw new Error("Column "+ columnName + " already exists in table " + getName());
            }
        }
    }

    /**
     * returns number of columns presently in 'this' TableSchema
     *
     * @return number of columns presently in 'this' TableSchema
     */
    public int size() {
        return columns.size();
    }

    /**
     * @return true if this schema is not attached to a DBSchema
     */
    public boolean isDetached() {
        return parent != null;
    }

    // CONSTRUCTORS
    /**
     * create table schema with given schemaName and an empty list of columns
     *
     * @param name of new TableSchema
     */
    public TableSchema(String name) {
        this.name = makeFirstLowerCase(name);
        columns = new ArrayList<>();
        parent = null;    // set by DBSchema.addTableSchema()
    }

    /**
     * create table schema with given schemaName and given set of columns
     *
     * @param name of new TableScheam
     * @param cols is list of column declarations
     */
    public TableSchema(String name, List<Column> cols) {
        this.name = name;
        columns = new ArrayList<>();
        columns.addAll(cols);
        parent = null;    // set by DBSchema.addTableSchema()
    }

    /**
     * copies table schema (this) for self-joins. The difference between a copy
     * and the original the newTableName. Names of columns are unchanged
     *
     * @param newTableName -- string to newTableName to give a different name to
     * copy
     * @return copied and renamed table schema
     */
    public TableSchema CopyForSelfJoins(String newTableName) {
        TableSchema dup = new TableSchema(newTableName, columns);
        return dup;
    }

    /**
     * reads prolog schema from File schemafileName that contains a single table
     * declaration; ParseExceptions and Errors are thrown
     *
     * read only .schema and database files
     *
     * @param schemafileName name of the file that contains a single Table
     * declaration
     * @return TableSchema of that file
     */
    public static TableSchema readSchema(String schemafileName) {
        try {
            File s = new File(schemafileName);
            return readSchema(s);
        } catch (Exception e) {
            throw new Error(e.getMessage());
        }
    }

    /**
     * reads prolog schema from File schemafile; errors are reported to out
     *
     * read only .schema and database files
     *
     * @param schemafile Java File of schema file to read (can also be a
     * database file, as it too embeds schema information)
     * @return DBSchema object
     */
    public static TableSchema readSchema(File schemafile) {
        TableSchema ts = null;

        String filename = schemafile.getName();

        if (!schemafile.exists()) {
            throw new Error(filename + " does not exist");
        }
        if (filename.contains(".ooschema.")) {
            throw new Error(filename + " is believed to be a .ooschema file");
        }

        boolean parsedTable = false;
        try {
            LineNumberReader br = new LineNumberReader(new InputStreamReader(new FileInputStream(schemafile)));
            String line;
            while ((line = br.readLine()) != null) {
                line = line.trim();
                if (line.equals("") || line.startsWith("%")) {
                    continue;
                }
                int lineno = br.getLineNumber();

                try {
                    if (line.startsWith("table")) {
                        if (parsedTable) {
                            throw new ParseException(lineno, "only one table definition permitted in " + filename);
                        }
                        ts = parseTableDecl(line, lineno);
                    }
                    // ignore everything else
                } catch (Exception e) {
                    throw new ParseException(lineno, "unable to parse table");
                }
            }
            br.close();

        } catch (IOException e) {
            throw new Error(e.getMessage());
        }
        if (ts == null) {
            throw new Error("returned table schema is null");
        }
        return ts;
    }

    /**
     * form the cross product of 'this' schema with otherSchema; this is a
     * concatenation of the column lists and a manufactured table name
     *
     * @param otherSchema -- to form cross product
     * @return composite schema
     */
    public TableSchema crossProduct(TableSchema otherSchema) {
        TableSchema newts = new TableSchema(name + "_x_" + otherSchema.getName());
        newts.addKols(this);
        newts.addKols(otherSchema);
        return newts;
    }

    /**
     * add the columns of ts to "this" tableschema. The names of the column go
     * from x to <ts.getName()>_x
     *
     * @param newts
     * @param ts
     */
    private void addKols(TableSchema ts) {
        String namePrefix;

        String schemaName = ts.getName();
        namePrefix = (schemaName.contains("_x_")) ? "" : schemaName + ".";

        for (Column c : ts.getColumns()) {
            String newColumnName = namePrefix + c.getName();
            this.addColumn(new Column(newColumnName, c.isQuoted()));
        }
    }

    // CONSTRUCTOR UTILITIES
    /**
     * add a new column c to 'this' TableSchema; new column is appended to the
     * existing list of columns
     *
     * Error is thrown if columns with duplicate names are detected
     *
     * @param c is column to add
     */
    public void addColumn(Column c) {
        doesNotHaveColumnEH(c.getName());
        columns.add(c); // append
    }

    /**
     * add column c as the first of 'this' schema
     *
     * @param c column to add
     */
    public void addColumnFirst(Column c) {
        doesNotHaveColumnEH(c.getName());
        columns.add(0, c); // put at front
    }

//    /**
//     * add a list of columns, columnList, to 'this' schema; if append is true,
//     * columnList is added to the end of 'this' column list; if false, it is
//     * added to the front
//     *
//     * @param columnList is list of columns to add
//     * @param append is true if columnList is to be appended
//     */
//    public void addColumns(List<Column> columnList) {
//        for (Column k : columnList) {
//            addColumn(k);
//        }
//    }

    /**
     * append an array of column to existing columns of 'this' schema
     *
     * @param columns is array of columns to add
     */
    public void addColumns(Column... columns) {
        for (Column k : columns) {
            addColumn(k);
        }
    }
    
    /**
     * place an array of columns before existing columns of 'this' schema
     *
     * @param columns is array of columns to add
     */
    public void addColumnsFirst(Column... columns) {
        // reverse the order of columns to get the correct insertion order
        Collections.reverse(Arrays.asList(columns)); 
        for (Column k : columns) {
            addColumnFirst(k);
        }
    }

    // UTILITIES
    /**
     * asserts if 'this' schema == otherSchema If false, Error is thrown
     *
     * @param otherSchema -- the schema to compare to
     */
    public void equalsEH(TableSchema otherSchema) {
        if (otherSchema == null) {
            throw new Error("argument schema is null for equalsEH");
        }
        if (!name.equals(otherSchema.getName())) {
            throw new Error("tables " + getName() + " and " + otherSchema.getName() + " do not have same schema names");
        }
        if (columns.size() != otherSchema.columns.size()) {
            throw new Error(getName() + " tables do not have same number of columns");
        }
        for (Column c : columns) {
            Column k = otherSchema.getColumn(c.getName());
            if (!c.equals(k)) {
                throw new Error("tables do not share column " + c.getName());
            }
        }
    }

    /**
     * returns true if 'this' schema equals otherSchema, false otherwise
     *
     * @param otherSchema -- other schema to compare
     *
     * @return true if both schemas are equal
     */
    public boolean equals(TableSchema otherSchema) {
        if (otherSchema == null) {
            return false;
        }
        if (columns.size() != otherSchema.columns.size()) {
            return false; // don't have the same # of columns
        }
        if (!name.equals(otherSchema.getName())) {
            return false; // dont have the same schema names
        }
        for (Column c : columns) {
            Column k = otherSchema.getColumn(c.getName());
            if (!c.equals(k)) {
                return false;  // no corresponding attribute
            }
        }
        return true;  // they match
    }

    // GET UTILITIES
    /**
     * @param name of column to findLocal
     * @return column with schemaName; Error thrown if not found
     */
    public Column getColumnEH(String name) {
        Column c = getColumn(name);
        if (c == null) {
            throw new Error("can't find column " + name + " in table " + this.name);
        } else {
            return c;
        }
    }

    /**
     *
     * @param name of column
     * @return column with schemaName; null if not found
     */
    public Column getColumn(String name) {
        for (Column c : columns) {
            if (c.getName().equals(name)) {
                return c;
            }
        }
        return null;
    }

    // PRINT 
    /**
     * returns string declaration of a TableSchema
     *
     * @return string declaration of a TableSchema
     */
    public String toString() {
        String q;
        String tb = "table(" + name + ",[";
        String comma = "";
        for (Column c : columns) {
            q = c.isQuoted() ? "\"" : "";
            tb = tb + comma + q + c.getName() + q;
            comma = ",";
        }
        tb = tb + "]).\n";
        return tb;
    }

    /**
     * print 'this' TableSchema to PrintStream out
     *
     * @param out where to print
     */
    public void print(PrintStream out) {
        out.print(this);
    }

    private static String makeFirstLowerCase(String s) {
        String tail = s.substring(1);
        String head = s.substring(0, 1).toLowerCase();
        return head + tail;
    }

}
