package PrologDB;

import MDELite6.Error;
import MDELite6.ParseException;
import PrologParser.LineToParse;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.function.Function;

/**
 * a tuple is an instance of a TableSchema
 */
public class Tuple {

    /**
     * Every tuple points to its TableSchema of origin this field should be
     * ignored once a table is computed
     *
     */
    private final TableSchema schema;
    
    /**
     * tuple contents are a map that pairs (column name, value)
     */
    private final Map<String, String> values;

    // ACCESSORS AND INFORMATION
    /**
     * returns value of given columnName; if columnName unrecognized, Error is
     * thrown
     *
     * Note: do not rename this method; required by velocity
     *
     * @param columnName -- name of column to read
     * @return value of given columnName
     */
    public String get(String columnName) {
        if (values.containsKey(columnName)) {
            return values.get(columnName);
        } else {
            schema.hasColumnEH(columnName);
            throw new Error("tuple missing value for " + columnName);
        }
    }

    /**
     * return int value of column with columnName; no checking is done to know
     * if column value can be converted into an int
     *
     * @param columnName -- name of column to read
     * @return int value of column with columnName
     */
    public int getInt(String columnName) {
        return Integer.parseInt(get(columnName));
    }

    /**
     * return double value of a column with columnName; no checking is done to
     * know if column value can be converted into an double
     *
     * @param columnName -- name of column to read
     * @return double value of a column with columnName
     */
    public double getDouble(String columnName) {
        return Double.parseDouble(get(columnName));
    }

    /**
     * return float value of a column with columnName; no checking is done to
     * know if column value can be converted into a float
     *
     * @param columnName -- name of column to read
     * @return float value of a column with columnName
     */
    public float getFloat(String columnName) {
        return Float.parseFloat(get(columnName));
    }

    /**
     * return boolean value of a column with columnName; no checking is done to
     * know if column value can be converted into a boolean
     *
     * @param columnName -- name of column to read
     * @return boolean value of a column (with columnName)
     */
    public boolean getBool(String columnName) {
        return Boolean.parseBoolean(get(columnName));
    }

    /**
     * returns 'null' if string content of columnName is "null" otherwise string
     * column value is returned
     *
     * @param columnName -- name of column to read
     * @return 'null' if string content of columnName is "null" otherwise string
     * column value is returned
     */
    public String getNull(String columnName) {
        String result = get(columnName);
        return result.equals("null") ? null : result;
    }

    /**
     * is value of columnName "null"?
     *
     * @param columnName -- name of column to check
     * @return true if null value
     */
    public boolean isNull(String columnName) {
        String result = get(columnName);
        return result.equals("null");
    }

    /**
     * does columnName have value?
     *
     * @param columnName -- name of column to check
     * @param value -- value column should have
     * @return true if column has this value
     */
    public boolean is(String columnName, String value) {
        String result = get(columnName);
        return result.equals(value);
    }

    /**
     * does columnName have any of these values
     * @param columnName -- name of column to check
     * @param values -- set of values the column might have
     * @return does valueOf(colunnName) in {values}?
     */
    public boolean is(String columnName, String... values) {
        String result = get(columnName);
        for (String s : values) {
            if (result.equals(s))
                return true;
        }
        return false;
    }

    /**
     * does columnName have integer value?
     *
     * @param columnName -- name of column to check
     * @param value -- integer value column should have
     * @return true if column has this value
     */
    public boolean is(String columnName, int value) {
        int result = getInt(columnName);
        return result == value;
    }

    /**
     * does columnName have float value?
     *
     * @param columnName -- name of column to check
     * @param value -- float value column should have
     * @return true if column has this value
     */
    public boolean is(String columnName, float value) {
        float result = getFloat(columnName);
        return result == value;
    }

    /**
     * does columnName have double value?
     *
     * @param columnName -- name of column to check
     * @param value -- double value column should have
     * @return true if column has this value
     */
    public boolean is(String columnName, double value) {
        double result = getDouble(columnName);
        return result == value;
    }

    /**
     * does columnName have boolean true?
     *
     * @param columnName -- name of column to check
     * @return true if column has true value
     */
    public boolean is(String columnName) {
        boolean result = getBool(columnName);
        return result;
    }

    /**
     * sets a column with columnName to have the given string value
     *
     * @param columnName -- name of column to update
     * @param value -- value to replace
     */
    public void set(String columnName, String value) {
        schema.hasColumnEH(columnName);
        values.put(columnName, value);
    }

    /**
     * sets a column with columnName to have the given int value
     *
     * @param columnName -- name of column to update
     * @param value -- value to replace
     */
    public void set(String columnName, int value) {
        schema.hasColumnEH(columnName);
        values.put(columnName, value + "");
    }

    /**
     * sets a column with columnName to have the given double value
     *
     * @param columnName -- name of column to update
     * @param value -- value to replace
     */
    public void set(String columnName, double value) {
        schema.hasColumnEH(columnName);
        values.put(columnName, value + "");
    }

    /**
     * sets a column with columnName to have the given float value
     *
     * @param columnName -- name of column to update
     * @param value -- value to replace
     */
    public void set(String columnName, float value) {
        schema.hasColumnEH(columnName);
        values.put(columnName, value + "");
    }

    /**
     * sets a column with columnName to have the given boolean value
     *
     * @param columnName -- name of column to update
     * @param value -- value to replace
     */
    public void set(String columnName, boolean value) {
        schema.hasColumnEH(columnName);
        values.put(columnName, value + "");
    }

    /**
     * sets a column with columnName to have null value
     *
     * @param columnName -- name of column to update
     */
    public void setNull(String columnName) {
        schema.hasColumnEH(columnName);
        values.put(columnName, null);
    }

    /**
     * returns TableSchema of 'this' tuple
     *
     * @return TableSchema of 'this' tuple
     */
    public TableSchema getSchema() {
        return schema;
    }

    /**
     * returns current number of (column,value) pairs in this tuple. When tuple
     * isComplete(), this should be the same number as the number of columns in
     * its tuple schema
     *
     * @return current number of (column,value) pairs in this tuple
     */
    public int size() {
        return values.size();
    }

    /**
     * returns list of all columns for this tuple.
     *
     * Method delegates to TableSchema.getColumns()
     *
     * @return list of all columns for this tuple
     */
    public List<Column> getColumns() {
        return schema.getColumns();
    }

    /**
     * projects 'this' tuple to the set of columns of a super table schema; if
     * table schemas do not match, tuple.getName() throws an Error projection
     * retains the first set of columns -- columns that are removed are at the
     * end of a tuple
     *
     * @param superTableSchema -- schema of supertable to project to
     * @return superTableSchema projected tuple
     */
    public Tuple project(TableSchema superTableSchema) {
        Tuple t = new Tuple(superTableSchema);
        for (Column c : superTableSchema.getColumns()) {
            String colName = c.getName();
            String value = get(colName);
            t.addColumnValue(colName, value);
        }
        t.isComplete();  // just to make sure
        return t;
    }

    /**
     * asserts that a tuple has its full complement of values. If not, an Error
     * is thrown
     */
    public void isComplete() {
        String errors = "";

        // Step 1: make sure that the tuple has AT LEAST its required set of columns
        for (Column c : schema.getColumns()) {
            String columnName = c.getName();
            if (!values.containsKey(columnName)) {
                errors += " " + columnName;
            }
        }
        if (!errors.equals("")) {
            throw new Error("tuple of " + schema.getName() + " missing values for columns" + errors);
        }

        // Step 2: make sure that the tuple has NO MORE than its required set of columns
        for (String colName : values.keySet()) {
            if (schema.getColumn(colName) == null) {
                errors += " " + colName;
            }
        }
        if (!errors.equals("")) {
            throw new Error("tuple of " + schema.getName() + " has too many columns " + errors);
        }
    }

    // CONSTRUCTORS
    /**
     * create an empty tuple for the given schema
     *
     * @param schema tuple schema to instantiate
     */
    public Tuple(TableSchema schema) {
        this.schema = schema;
        values = new HashMap<>();
    }

    /**
     * create an empty tuple for the given table
     *
     * @param table -- table to which tuple should conform
     */
    public Tuple(Table table) {
        this(table.getSchema());
    }

    /**
     * returns a (deep) copy of a tuple;
     * precondition: newSchema and tableSchema must have identically
     * defined sets of columns; the schemas may have different names
     *
     * @param newSchema -- returned tuple copy must be instance of this schema
     * @return -- copy of tuple
     */
    public Tuple copy(TableSchema newSchema) {
        Tuple cpy = new Tuple(newSchema);
        cpy.values.putAll(values);
        return cpy;
    }

    /**
     * Add the (column,value) pairs of tuple t to the existing tuple.
     * @param t -- tuple whose (column, value) pairs are to be added to tuple 'this'
     * used in table joins; not exported outside of package
     */
    void addColumnValues(Tuple t) {
        String tableSchemaName = t.getSchema().getName();
        if (tableSchemaName.contains(".") || tableSchemaName.contains("_x_")) {
            tableSchemaName = "";
        } else {
            tableSchemaName = tableSchemaName + ".";
        }

        for (Entry<String, String> entry : t.values.entrySet()) {
            String colName = tableSchemaName + entry.getKey();
            this.addColumnValue(colName, entry.getValue());
        }
    }


    // CONSTRUCTOR UTILITIES
    /**
     * adds a list of values, in order in which their columns are defined, to an
     * empty tuple. an Error is thrown if not all columns have values.
     *
     * @param vals is list of values to add to an empty tuple
     */
    public void addColumnValuesEH(String... vals) {
        if (vals.length != schema.size()) {
            throw new Error("adding " + vals.length + " column values to table " + schema.getName() + " with " + schema.size() + "columns");
        }
        int i = 0;
        for (Column c : schema.getColumns()) {
            values.put(c.getName(), vals[i]);
            i++;
        }
        isComplete();
    }

    /**
     * adds an individual column,value pair to a tuple; Error thrown if column
     * is not existent
     *
     * @param name of column
     * @param value value of column
     */
    public void addColumnValue(String name, String value) {
        schema.getColumnEH(name);  // throws error if non-existent
        values.put(name, value);
    }
    
    /**
     * parses a line at lineno, expecting to parse a legal prolog tuple
     * declaration that conforms to table schema ts
     *
     * @param line to parse
     * @param lineno line number of the line
     * @param ts table schema for tuple to conform
     * @return Tuple
     *
     */
    public static Tuple parseTupleDecl(String line, int lineno, TableSchema ts) {
        // Step 1: parse tuple
        LineToParse ds = new LineToParse(line, lineno);
        ds.parser(LineToParse.tupleStmt);  // ignore output

        // Step 2: unpack parsing to get table name, and list of values, and create an empty tuple
        LinkedList<String> list = ds.parseList;
        String tableName = list.removeFirst();  // this is the table name
        Tuple newTuple = new Tuple(ts);
        if (!tableName.equals(ts.getName())) {
            throw new ParseException(lineno, "tuple name does not equal table name");
        }

        if (list.size() != ts.size()) {
            throw new ParseException(lineno, "insufficient number of values for a tuple of table " + ts.getName());
        }
        // Step 3: assemble the tuple
        for (int i = 0; i < ts.size(); i++) {
            // Get the value delimited by "@@@"
            String value = list.removeFirst();  // this is a column value
            value = value.replace("'", "");     // remove single quotes
            Column target = ts.getIndex(i);
            String columnName = target.getName().replace("\"", "");
            newTuple.addColumnValue(columnName, value);
        }

        // Step 4: return tuple
        return newTuple;
    }

    // PRINT 
    /**
     * print tuple to PrintStream out
     *
     * @param out where to print
     */
    public void print(PrintStream out) {
        out.print(this);
    }

    /**
     * converts tuple into a string; throws error if # of columns of tuple do
     * not match # of columns in table schema
     *
     * @return string of tuple
     */
    public String toString() {
        if (values.size() != schema.size()) {
            throw new Error("tuple from table " + schema.getName() + " does not have correct number of attributes");
        }
        String q;
        String tuple = schema.getName() + "(";
        String comma = "";
        for (Column c : schema.getColumns()) {
            q = c.isQuoted() ? "'" : "";
            String val = values.get(c.getName());
            if (val == null && !c.isQuoted()) {
                val = "null";
            }
            tuple = tuple + comma + q + val + q;
            comma = ",";
        }
        tuple = tuple + ").\n";
        return tuple;
    }

    // TUPLE JOIN Operations

    /**
     * returns the first tuple of otherTable that joins with 'this', where
     * "this".thisJoinkey == otherTable.otherJoinKey; returns null otherwise
     *
     * @param otherTable -- the other table to which 'this' is to be joined
     * @param thisJoinKey -- the join key of 'this' table
     * @param otherJoinKey -- the join key of otherTable
     * @return -- the first tuple of otherTable that joins with "this", null
     * otherwise
     */
    public Tuple joinFirst(String thisJoinKey, Table otherTable, String otherJoinKey) {
        String value = get(thisJoinKey);
        for (Tuple tb : otherTable.tuples()) {
            if (tb.is(otherJoinKey, value)) {
                return tb;
            }
        }
        return null;
    }

    /**
     * same as firstJoin, except an Error is thrown if no tuple of otherTable is
     * found.
     *
     * @param otherTable -- the other table to which 'this' is to be joined
     * @param thisJoinKey -- the join key of 'this' table
     * @param otherJoinKey -- the join key of otherTable
     * @return -- the first tuple of otherTable that joins with "this"
     */
    public Tuple joinFirstEH(String thisJoinKey, Table otherTable, String otherJoinKey) {
        Tuple result = joinFirst(thisJoinKey, otherTable, otherJoinKey);
        if (result == null) {
            throw new Error("no tuple in " + otherTable.getName() + " found with key " + get(thisJoinKey));
        }
        return result;
    }
    
    /**
     * transform a Tuple to an object of type S
     *
     * @param <S> -- return type, should not be Tuple or subclass of Tuple
     * @param f -- a function that maps a Table to S
     * @return -- object of type S
     */
    public <S> S map(Function<Tuple,S> f) {
        return f.apply(this);
    }
}
