package PrologParser;

import MDELite6.ParseException;
import PrologProductions.Name;
import PrologProductions.LeftParen;
import PrologProductions.LeftBracket;
import PrologProductions.Fixed;
import PrologProductions.NameList;
import PrologProductions.RightParen;
import PrologProductions.TupleValues;
import PrologProductions.Dot;
import PrologProductions.End;
import PrologProductions.RightBracket;
import PrologProductions.Comma;
import PrologProductions.ColumnDecl;
import PrologProductions.Token;
import java.util.LinkedList;

/**
 * a LineToParse is a line to be parsed (the String of a line + its line number)
 */
public class LineToParse extends PrimitiveLookAhead {
    
    /**
     * dbase statement as a sequence of tokens
     */
    static public Token[] dbaseStmt
            = {new Fixed("dbase"), new LeftParen(), new Name(), new Comma(), new LeftBracket(), new NameList(), new RightBracket(), new RightParen(), new Dot(), new End()};
    
    /**
     * table statement as a sequence of tokens
     */
    static public Token[] tableStmt
            = {new Fixed("table"), new LeftParen(), new Name(), new Comma(), new LeftBracket(), new ColumnDecl(), new RightBracket(),
                new RightParen(), new Dot(), new End()};
    
    /**
     * tuple statement as a sequence of tokens
     */
    static public Token[] tupleStmt
            = {new Name(), new LeftParen(), new TupleValues(), new RightParen(), new Dot(), new End()};
    
    /**
     * subTableStatement statement as a sequence of tokens
     */
    static public Token[] subTableStmt
            = {new Fixed("subtable"), new LeftParen(), new Name(), new Comma(), new LeftBracket(), new NameList(), new RightBracket(),
                new RightParen(), new Dot(), new End()};

    /**
     * toParse is the string that is to be parsed
     */
    public String toParse;
    
    /**
     * token is the most recently parsed (and to be remembered) token
     * Some tokens (whose values we know) need not be remembered
     */
    public String token;
    
    /**
     * parselist is a linkedList of tokens (strings) that capture the essence of a parse
     */
    public LinkedList<String> parseList;
    
    /**
     * lineno is the line number of the string, toParse
     */
    public int lineno;

    /**
     * A phrase is really a lineContents to be parsed, defined by a String 
     * lineContents and its lineNumber
     * @param lineContents  -- string contents of a line
     * @param lineNumber -- line number associated with this string
     */
    public LineToParse(String lineContents, int lineNumber) {
        toParse = lineContents;
        token = null;
        parseList = new LinkedList<>();
        lineno = lineNumber;
    }

    /**
     * look at next character to parse
     * @return next character to parse
     */
    public char peek() {
        return toParse.charAt(0);
    }
    
    /**
     * is the next character k?
     * @param k -- ideally the next character to parse
     * @return true if k is next char to parse
     */

    public boolean parseChar(char k) {
        if (toParse.equals("")) {
            return false;
        }
        return (peek() == k) ? returnResult(1) : false;
    }

    /**
     * is there more to parse to parse? If not, throw ParseException
     */
    public void AssertThereIsMoreToParse() {
        if (toParse.equals("")) {
            throw new ParseException(lineno, "End Of String Reached");
        }
    }

    /**
     * advance toParse by the length i of a parsed "token" and
     * set token to the length i string that has been removed;
     * this is the "result" of parsing.  Return true.
     * 
     * @param i -- length of parsed "token" string
     * @return true
     */
    public boolean returnResult(int i) {
        token = toParse.substring(0, i);
        toParse = toParse.substring(i).trim();
        return true;
    }

    /**
     * advance toParse by the length i of a parsed "token"; no 
     * token is returned; return true.
     * 
     * @param i -- length of parsed token
     * @return true
     */
    public boolean returnSkip(int i) {
        toParse = toParse.substring(i).trim();
        return true;
    }

    /**
     * if b is false (meaning that the requested parsing step cannot
     * be done), throw a ParseException with the given error message emsg;
     * otherwise add the parsed token to the parseList.
     * 
     * @param b -- boolean whether requested parsing step can be done
     * @param emsg -- error message given when b is false
     */
    public void parseStep(boolean b, String emsg) {
        if (b) {
            parseList.add(token);
        } else {
            throw new ParseException(lineno, emsg + " expected at >>" + toParse);
        }
    }

    /**
     * if b is false (meaning that the requested parsing step cannot 
     * be done), throw a ParseException with the given error message emsg;
     * otherwise do nothing (nothing to remember on parseList).
     * @param b -- boolean whether requested parsing step can be done
     * @param emsg -- error message given when b is false
     */
    public void parseSkip(boolean b, String emsg) {
        if (b) {
        } else {
            throw new ParseException(lineno, emsg + " expected at >>" + toParse);
        }
    }

    /**
     * parse a statement (a sequence of tokens); return true
     * @param statement -- sequence of tokens/statement expected
     * @return -- true (if no ParseExceptions are thrown)
     */
    public boolean parser(Token[] statement) {
        for (Token t : statement) {
            t.setLineToParse(this);
            t.parse();
        }
        return true;
    }

}
