package PrologParser;

import MDELite6.ParseException;
import MDELite6.Error;
import PrologDB.Column;
import PrologDB.DB;
import PrologDB.DBSchema;
import PrologDB.SubTableSchema;
import PrologDB.Table;
import PrologDB.TableSchema;
import PrologDB.Tuple;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;


/**
 * this is a simple line-parser for database and schema specifications
 * Generally this class is internal to PrologDB package and should not
 * be called externally
 */
public class Parser {

    /**
     * turn Debug on if you want to see stack trace on Errors
     */
    static boolean Debug = false;

    /**
     * a dbase declaration is on a single line. this method parses the given
     * line @ lineno returning the DBSchema given declared name, also returns a
     * list of tableNames (not their schemas).  ParseExceptions thrown if
     * unable to parse
     *
     * @param line -- a string
     * @param lineno -- the line number associated with this line
     * @param tableNames -- an empty list which is to be filled with table names
     * @return DBSchema object with name of schema initialized; also returns
     * list of tableNames -- to be used later in parsing
     */
    public static DBSchema parseDBaseDecl(String line, int lineno, List<String> tableNames) {
        // Step 1: parse the line
        LineToParse ds = new LineToParse(line, lineno);
        ds.parser(LineToParse.dbaseStmt);

        // Step 2: unpack parsing
        LinkedList<String> list = ds.parseList;
        String schemaName = list.removeFirst();  // is name of the dbase schema
        DBSchema dbs = new DBSchema(schemaName);
        tableNames.addAll(list);
        return dbs;
    }

    /**
     * a subtable declaration is on a single line. this method parses the given line 
     * at lineno and adds the subtable declaration to DBSchema dbs. A ParseException is
     * thrown if something wrong (parse or semantic error) is discovered.
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add subtable definition
     */
    public static void parseSubTableDecl(String line, int lineno, DBSchema dbs) {
        if (dbs == null) {
            throw new ParseException(lineno,"subtable defined before dbase");
        }

        // Step 1: parse the line
        LineToParse ds = new LineToParse(line, lineno);
        ds.parser(LineToParse.subTableStmt);

        // Step 2: unpack parsing to get table name, and list of values, and create an empty tuple
        LinkedList<String> list = ds.parseList;
        String tableName = list.removeFirst();  // this is the super table name
        TableSchema suptab = dbs.getTableSchemaEH(tableName);
        ArrayList<TableSchema> subTableSchemas = new ArrayList<>();
        for (String name : list) {
            TableSchema ts = dbs.getTableSchemaEH(name);
            subTableSchemas.add(ts);
        }

        // Step 3: assemble the subtableschema declaaration
        SubTableSchema sts = new SubTableSchema(suptab, subTableSchemas);
        dbs.addSubTableSchema(sts);
    }

    /**
     * a table declaration is on a single line. this method parses the given
     * line @ lineno and adds the table definition to schema dbs
     *
     * @param line to parse
     * @param lineno of line
     * @param dbs to add table declaration
     * @throws Error -- when parsing errors occurs
     */
    public static void parseTableDecl(String line, int lineno, DBSchema dbs) throws Error {
        // Step 0: simple error checking;
        if (dbs == null) {
            throw new Error("table defined before dbase");
        }
        
        TableSchema ts = parseTableDecl(line,lineno);

        dbs.addTableSchema(ts);
    }
    
    /**
     * a table declaration is on a single line. this method parses the given
     * line @ lineno; returns tableschema
     *
     * @param line to parse
     * @param lineno of line
     * @throws Error -- when parsing errors occurs
     * @return TableSchema of line input
     */
    public static TableSchema parseTableDecl(String line, int lineno) throws Error {

        // Step 1: parse line
        LineToParse ds = new LineToParse(line, lineno);
        ds.parser(LineToParse.tableStmt);

        // Step 2: unpack parsing to get table name, and list of values, and create an empty tuple
        LinkedList<String> list = ds.parseList;
        String tableName = list.removeFirst();  // this is the table name
        TableSchema tableDef = new TableSchema(tableName);

        // Step 3: assemble the column defs
        for (String name : list) {
            boolean isQuoted = name.startsWith("\"");
            Column newColumn = new Column(name.replace("\"", ""), isQuoted);
            tableDef.addColumn(newColumn);
        }
        
        // Step 4: return tableDef
        return tableDef;
    }

    /**
     * parses a line at lineno, expecting to parse a legal prolog tuple
     * declaration. The extracted tuple is inserted into database db; the
     * possible table declarations are defined in db.getSchema() of the database
     *
     * @param line to parse
     * @param lineno line number of the line
     * @param db database in which to insert
     *
     */
    protected static void parseTupleDecl(String line, int lineno, DB db) {
        // Step 1: parse tuple
        LineToParse ds = new LineToParse(line, lineno);
        ds.parser(LineToParse.tupleStmt); // ignore output
        //Step 2: extract the database schema
        DBSchema dbs = db.getSchema();
        // Step 2: unpack parsing to get table name, and list of values, and create an empty tuple
        LinkedList<String> list = ds.parseList;
        String tableName = list.removeFirst(); // this is the table name
        Table tbl = db.getTableEH(tableName);
        TableSchema tableDef = db.getTableSchema(tableName);
        Tuple newTuple = new Tuple(tableDef);
        if (list.size() != tableDef.size()) {
            throw new Error("at line " + lineno + " insufficient number of values for a tuple of table " + tableDef.getName());
        }
        // Step 3: assemble the tuple
        for (int i = 0; i < tableDef.size(); i++) {
            // Get the value delimited by "@@@"
            String value = list.removeFirst(); // this is a column value
            value = value.replace("'", ""); // remove single quotes
            Column target = tableDef.getIndex(i);
            String columnName = target.getName().replace("\"", "");
            newTuple.addColumnValue(columnName, value);
        }
        // Step 4: add the tuple to the table
        tbl.add(newTuple);
    }

}
