package vm2t;

import MDELite6.Constants;
import MDL.Vm2t;
import PrologDB.DB;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.io.StringReader;
import java.io.StringWriter;
import java.lang.reflect.Constructor;
import java.util.Properties;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.Velocity;


/**
 * this is the main engine that integrates PrologDB facilities 
 * with velocity;  @see MDL.Vm2t for its facade
 */
public class Main {

    static String prologFileName;
    static String templateFileName;
    static String outputFileName;
    static String contextGenClassName;

    
    static File openEH(String errMsg, String filename) {
        
        File fl = new File(filename);

            if (!fl.exists()) {
                throw new RuntimeException( errMsg + templateFileName);
            }
        return fl;
    }

    /**
     * main engine that integrates PrologDB with Velocity
     * @see MDL.Vm2t for its facade.
     * @param args  -- see MDL.Vm2t for its facade.
     */
    public static void main(String... args) {

        prologFileName = null;
        templateFileName = null;
        outputFileName = null;
        contextGenClassName = null;

        // Step 1: read the command line arguments
        switch (args.length) {
            case 2: /* prolog,template */

                prologFileName = args[0];
                templateFileName = args[1];
                outputFileName = Constants.vm2tFile;
                break;
            case 3: /* prolog, template, out */

                prologFileName = args[0];
                templateFileName = args[1];
                outputFileName = args[2];
                break;
            case 4: /* prolog, template, -cg, context */

                if (args[2].equals("-cg")) {
                    Vm2t.marquee();
                    return;
                }
                prologFileName = args[0];
                templateFileName = args[1];
                outputFileName = Constants.vm2tFile;
                contextGenClassName = args[3];
                break;
            case 5: /* prolog, template, out, cg, context */

                if (args[3].equals("-cg")) {
                    Vm2t.marquee();
                    return;
                }
                prologFileName = args[0];
                templateFileName = args[1];
                outputFileName = args[2];
                contextGenClassName = args[4];
                break;
            default:
                Vm2t.marquee();
                return;
        }

        // Step 2: now execute the template file
        try {
            File prologFile = openEH("prolog-file does not exist: ",prologFileName);
            // the next line fixes a problem with Velocity -- it strips off white space from lines
            // that have # in them.  this allows sane indenting
            File tf = openEH("template-file does not exist: ",templateFileName);
            File templateFile = TemplateFile.cutWhiteSpace(tf);
            //File fl = new File(templateFileName);


            Properties p = new Properties();
            String templatePath = ".";
            if (templateFile.getParentFile() != null) {
                templatePath = templateFile.getParentFile().getCanonicalPath();
            }
            p.setProperty("file.resource.loader.path", templatePath);
            Velocity.setProperty(Velocity.RUNTIME_LOG_LOGSYSTEM, new CustomLog());
//            Velocity.setProperty(Velocity.INPUT_ENCODING, "UTF-8");
//            Velocity.setProperty(Velocity.OUTPUT_ENCODING, "UTF-8");
            Velocity.init(p);

            DB m =  DB.readDataBase(prologFile, System.err);

            VelocityContext context;
            if (contextGenClassName != null) {
                @SuppressWarnings("unchecked")
                Class<ContextGenerator> cgClass = (Class<ContextGenerator>) Class.forName(contextGenClassName);
                Constructor<ContextGenerator> cgCons = cgClass.getConstructor();
                ContextGenerator cgIns = cgCons.newInstance();
                context = cgIns.generateContext(prologFileName, m, outputFileName);
            } else {
                context = new DefaultContextGenerator().generateContext(prologFileName, m, outputFileName);
            }
            context.put("NEWLINE","\n");
            
            Template template;
            template = Velocity.getTemplate(templateFile.getName());
            StringWriter writer = new StringWriter();

            template.merge(context, writer);
            if (context.get("MARKER") == null) {
                System.out.println(writer.toString());  // so if no MARKER is declared output to std out
            } else {
                String deldir = null;
                Object o = context.get("DELDIR");
                if (o != null) {
                    deldir = o.toString();
                }
                splitter(writer.toString(), context.get("MARKER").toString(), deldir);
            }

            //templateFile.delete();
        } catch (ClassNotFoundException e) {
            System.err.println("The context generator class not found. Please make "
                    + "sure it is in the java classpath.");
            System.err.close();
        } catch (Exception e) {
            System.err.println(e.getMessage());
            System.err.close();
        }

    }

    static void splitter(String content, String marker, String deldir) {
        BufferedReader br = new BufferedReader(new StringReader(content));
        String line;
        PrintStream out = null;
        String filename, dirname;

        try {
            while ((line = br.readLine()) != null) {
                if (line.startsWith(marker)) {
                    if (out != null) {
                        out.close();
                    }
                    filename = line.substring(marker.length());
                    File f = new File(filename);
                    if (f.getParentFile() != null && !f.getParentFile().exists()) {
                        f.getParentFile().mkdirs();
                    }
                    out = new PrintStream(f);
                } else if (deldir != null && line.startsWith(deldir)) {
                    dirname = line.substring(deldir.length());
//                    System.out.println("dirname = "+ dirname);
                    File f = new File(dirname);
                    if (f.exists() && f.isDirectory()) {
                        recursiveDelete(f);
                    }
                } else {
                    if (out != null) {
                        out.println(line);
                    }
                }
            }
        } catch (IOException e) {
            System.err.println(e.getMessage());
            if (out != null) {
                out.close();
            }
        }
        if (out != null) {  // lines added because the last file was never closed
            out.close();
        }
    }

    // modified from http://www.mkyong.com/java/how-to-recursiveDelete-directory-in-java/
    static void recursiveDelete(File file) throws IOException {

        if (file.isDirectory()) {
            //directory is empty, then recursiveDelete it
            if (file.list().length == 0) {
                file.delete();
            } else {
                //list all the directory contents
                String files[] = file.list();
                for (String temp : files) {
                    //construct the file structure
                    File fileDelete = new File(file, temp);
                    //recursive recursiveDelete
                    recursiveDelete(fileDelete);
                }

                //check the directory again, if empty then recursiveDelete it
                if (file.list().length == 0) {
                    file.delete();
                }
            }
        } else {
            //if file, then recursiveDelete it
            file.delete();
        }
    }
}
